@@ -1,5 +1,5 @@
 /*******************************************************************************
- * Copyright (c) 2006 IBM Corporation.
+ * Copyright (c) 2006 IBM Corporation and others.
  * All rights reserved. This program and the accompanying materials
  * are made available under the terms of the Eclipse Public License v1.0
  * which accompanies this distribution, and is available at
@@ -7,6 +7,7 @@
  *
  * Contributors:
  *     IBM Corporation - Jeff Briggs, Henry Hughes, Ryan Morse
+ *     Red Hat - ongoing maintenance
  *******************************************************************************/
 
 package org.eclipse.linuxtools.systemtap.structures.runnable;
@@ -69,7 +70,7 @@ public class Command implements Runnable {
      */
     protected final String[] envVars;
 
-    protected Process process;
+    protected Process process = null;
     /**
      * @since 2.1
      */
@@ -113,16 +114,20 @@ public class Command implements Runnable {
     /**
      * Starts the <code>Thread</code> that the new <code>Process</code> will run in.
      * This must be called in order to get the process to start running.
-     * @throws CoreException
+     * Note that this method only takes effect the first time it is called.
+     * @throws CoreException If initializing failed.
      */
     public void start() throws CoreException {
+        if (started || stopped) {
+            return;
+        }
         IStatus status = init();
         if (status.isOK()) {
             Thread t = new Thread(this, cmd[0]);
             t.start();
             started = true;
         } else {
-            stop();
+            cleanUpAfterStop();
             returnVal = Integer.MIN_VALUE;
             throw new CoreException(status);
         }
@@ -131,24 +136,22 @@ public class Command implements Runnable {
     /**
      * Starts up the process that will execute the provided command and registers
      * the <code>StreamGobblers</code> with their respective streams.
+     * @return The status from the initializing process
      * @since 2.0
      */
     protected IStatus init() {
         try {
             process = RuntimeProcessFactory.getFactory().exec(cmd, envVars, project);
-
-            if (process == null) {
-                return new Status(IStatus.ERROR, StructuresPlugin.PLUGIN_ID, Messages.Command_failedToRunSystemtap);
-            }
-
-            errorGobbler = new StreamGobbler(process.getErrorStream());
-            inputGobbler = new StreamGobbler(process.getInputStream());
-
-            this.transferListeners();
-            return Status.OK_STATUS;
         } catch (IOException e) {
             return new Status(IStatus.ERROR, StructuresPlugin.PLUGIN_ID, e.getMessage(), e);
         }
+        if (process == null) {
+            return new Status(IStatus.ERROR, StructuresPlugin.PLUGIN_ID, Messages.Command_failedToRunSystemtap);
+        }
+        errorGobbler = new StreamGobbler(process.getErrorStream());
+        inputGobbler = new StreamGobbler(process.getInputStream());
+        transferListeners();
+        return Status.OK_STATUS;
     }
 
     /**
@@ -169,7 +172,7 @@ public class Command implements Runnable {
     /**
      * This method handles checking the status of the running <code>Process</code>. It
      * is called when the new Thread is created, and thus should never be called by
-     * any implementing program. To run call the <code>start</code> method.
+     * any implementing program. To run call the {@link #start} method.
      */
     @Override
     public void run() {
@@ -177,16 +180,23 @@ public class Command implements Runnable {
         inputGobbler.start();
         try {
             process.waitFor();
-        } catch (InterruptedException e) {}
-        stop();
+            cleanUpAfterStop();
+        } catch (InterruptedException e) {
+            // This thread was interrupted while waiting for
+            // the process to exit. Destroy the process just
+            // to make sure it exits.
+            stop();
+        }
     }
 
     /**
-     * Stops the process from running and stops the <code>StreamGobblers</code> from monitering
-     * the dead process and unregisters the StreamListener. Also wakes up any threads waiting
-     * on this command.
+     * Performs cleanup operations for when the process ends:
+     * Stops the <code>StreamGobblers</code> from monitering
+     * the dead process and unregisters the StreamListener.
+     * Also wakes up any threads waiting on this command.
+     * @since 3.1
      */
-    public synchronized void stop() {
+    protected synchronized void cleanUpAfterStop() {
         if (!stopped) {
             if (errorGobbler != null) {
                 errorGobbler.stop();
@@ -194,22 +204,24 @@ public class Command implements Runnable {
             if (inputGobbler != null) {
                 inputGobbler.stop();
             }
-            try {
-                if (process != null) {
-                    process.waitFor();
-                }
-            } catch (InterruptedException e) {
-                // This thread was interrupted while waiting for
-                // the process to exit. Destroy the process just
-                // to make sure it exits.
-                process.destroy();
-            }
             removeInputStreamListener(logger);
             stopped = true;
             notifyAll(); // Wake up threads waiting for this command to stop.
         }
     }
 
+    /**
+     * Stops the process from running and performs post-stop cleanup if necessary.
+     */
+    public synchronized void stop() {
+        if (!stopped) {
+            if (process != null) {
+                process.destroy();
+            }
+            cleanUpAfterStop();
+        }
+    }
+
     /**
      * Method to check whether or not the process is running.
      * @return The execution status.
@@ -272,7 +284,7 @@ public class Command implements Runnable {
 
     /**
      * Removes the provided listener from those monitoring the InputStream.
-     * @param listener An </code>IGobblerListener</code> that is monitoring the stream.
+     * @param listener An <code>IGobblerListener</code> that is monitoring the stream.
      */
     public void removeInputStreamListener(IGobblerListener listener) {
         if (inputGobbler != null) {
@@ -284,7 +296,7 @@ public class Command implements Runnable {
 
     /**
      * Removes the provided listener from those monitoring the ErrorStream.
-     * @param listener An </code>IGobblerListener</code> that is monitoring the stream.
+     * @param listener An <code>IGobblerListener</code> that is monitoring the stream.
      */
     public void removeErrorStreamListener(IGobblerListener listener) {
         if (errorGobbler != null) {
@@ -295,9 +307,10 @@ public class Command implements Runnable {
     }
 
     /**
-     * Saves the input stream data to a permanent file.  Any new data on the
+     * Saves the input stream data to a permanent file. Any new data on the
      * stream will automatically be saved to the file.
      * @param file The file to save the InputStream to.
+     * @return True if the save was successful, false otherwise.
      */
     public boolean saveLog(File file) {
         return logger.saveLog(file);
@@ -307,15 +320,18 @@ public class Command implements Runnable {
      * Disposes of all internal components of this class. Nothing in the class should be
      * referenced after this is called.
      */
-    public void dispose() {
+    public synchronized void dispose() {
         if (!disposed) {
             stop();
-            disposed = true;
-
-            inputListeners.clear();
-            errorListeners.clear();
 
+            if (inputListeners != null) {
+                inputListeners.clear();
+            }
             inputListeners = null;
+
+            if (errorListeners != null) {
+                errorListeners.clear();
+            }
             errorListeners = null;
 
             if (inputGobbler != null) {
@@ -327,7 +343,10 @@ public class Command implements Runnable {
                 errorGobbler.dispose();
             }
             errorGobbler = null;
+
             logger.dispose();
+            process = null;
+            disposed = true;
         }
     }
 
